__all__ = [
    "Fragment",
    "InvalidFragment",
]


from base64 import b64decode
from dataclasses import dataclass, replace
from pathlib import Path
from typing import Any, Optional, Sequence, Type, TypeVar, overload
from urllib.parse import urlparse
from urllib.request import urlopen

from beet import BinaryFile, BinaryFileBase, BubbleException, Cache, File
from beet.core.utils import FileSystemPath

FileType = TypeVar("FileType", bound=File[Any, Any])


class InvalidFragment(BubbleException):
    """Raised when a fragment can not be processed."""

    message: str
    line: int

    def __init__(self, message: str, line: int):
        super().__init__(message, line)
        self.message = message
        self.line = line

    def __str__(self) -> str:
        return f"{self.message} (line {self.line + 1})"


@dataclass(frozen=True)
class Fragment:
    """Class representing a fragment annotated by a directive."""

    start_line: int
    end_line: int
    directive: str
    modifier: Optional[str] = None
    arguments: Sequence[str] = ()
    content: Optional[str] = None
    url: Optional[str] = None
    path: Optional[FileSystemPath] = None
    file: Optional[File[Any, Any]] = None
    cache: Optional[Cache] = None

    def with_content(self, content: str) -> "Fragment":
        """Replace content."""
        return replace(self, content=content)

    def with_link(
        self,
        link: Any,
        external_files: Optional[FileSystemPath] = None,
    ) -> "Fragment":
        """Replace linked content."""
        url = str(link)
        path = None

        if urlparse(url).path == url:
            if external_files:
                path = Path(external_files, url).resolve()
            url = None

        return replace(self, url=url, path=path)

    @overload
    def expect(self) -> None: ...

    @overload
    def expect(self, name1: str, /) -> str: ...

    @overload
    def expect(self, name1: str, name2: str, /, *names: str) -> Sequence[str]: ...

    def expect(self, *names: str):
        """Check directive arguments."""
        if missing := names[len(self.arguments) :]:
            msg = f"Missing argument {', '.join(map(repr, missing))} for directive @{self.directive}."
            raise InvalidFragment(msg, self.start_line)
        if extra := self.arguments[len(names) :]:
            msg = f"Unexpected argument {', '.join(map(repr, extra))} for directive @{self.directive}."
            raise InvalidFragment(msg, self.start_line)
        if len(self.arguments) == 0:
            return
        if len(self.arguments) == 1:
            return self.arguments[0]
        return self.arguments

    @overload
    def as_file(self) -> BinaryFile: ...

    @overload
    def as_file(self, file_type: Type[FileType]) -> FileType: ...

    def as_file(self, file_type: Type[File[Any, Any]] = BinaryFile) -> File[Any, Any]:
        """Retrieve the content of the fragment as a file."""
        is_binary = issubclass(file_type, BinaryFileBase)

        if self.file:
            content = self.file.ensure_serialized()
            if is_binary and isinstance(content, str):
                content = content.encode()
            elif not is_binary and isinstance(content, bytes):
                content = content.decode()
            return file_type(content)

        content = self.content

        if content is not None and self.modifier == "base64":
            content = b64decode(content.strip())

        elif content is not None and self.modifier == "download" or self.url:
            url = content.strip() if content is not None else self.url

            if not (url and url.startswith(("http:", "https:", "data:"))):
                raise InvalidFragment(f"Invalid url {url!r}.", self.start_line)

            if self.cache and not url.startswith("data:"):
                return file_type(source_path=self.cache.download(url))

            with urlopen(url) as f:
                content = f.read()

        elif content is not None:
            if self.modifier == "strip_final_newline" and content.endswith("\n"):
                content = content[:-1]

            return file_type(content.encode() if is_binary else content)

        elif self.path:
            return file_type(source_path=self.path)

        else:
            msg = f"Expected content, path or url for directive @{self.directive}."
            raise InvalidFragment(msg, self.start_line)

        return file_type(content if is_binary else content.decode(errors="replace"))



"""Plugin that handles yaml fragments for json files."""

__all__ = [
    "handle_yaml",
]


from dataclasses import replace
from typing import Mapping

import yaml
from beet import Context, JsonFile

from lectern import (
    DataPackDirective,
    Directive,
    Document,
    Fragment,
    NamespacedResourceDirective,
    ResourcePackDirective,
)


def beet_default(ctx: Context):
    document = ctx.inject(Document)
    document.loaders.append(handle_yaml)


def handle_yaml(fragment: Fragment, directives: Mapping[str, Directive]) -> Fragment:
    """Loader that converts yaml to json."""
    directive = directives.get(fragment.directive)
    is_yaml = False

    if isinstance(directive, NamespacedResourceDirective):
        is_yaml = directive.file_type.extension == ".json"

    elif isinstance(directive, (DataPackDirective, ResourcePackDirective)):
        relative_path = fragment.expect("relative_path")
        if is_yaml := relative_path.endswith((".yml", ".yaml")):
            fragment = replace(
                fragment,
                arguments=[
                    relative_path.replace(".yml", ".json").replace(".yaml", ".json")
                ],
            )

    if is_yaml:
        yaml_file = fragment.as_file(JsonFile)
        yaml_file.ensure_deserialized(yaml.safe_load)
        fragment = replace(fragment, file=yaml_file)

    return fragment



__all__ = [
    "Extractor",
    "TextExtractor",
    "EmbeddedExtractor",
    "MarkdownExtractor",
    "MarkdownParserWithUncheckedLinks",
    "FragmentLoader",
]


import json
import re
from dataclasses import replace
from itertools import islice
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Union,
)

from beet import Cache, DataPack, ResourcePack
from beet.core.utils import FileSystemPath
from markdown_it.main import MarkdownIt
from markdown_it.token import Token

from .directive import Directive
from .fragment import Fragment

FragmentLoader = Callable[[Fragment, Mapping[str, Directive]], Optional[Fragment]]


RELATIVE_PATH_REGEX = re.compile(
    r"^(?:[a-zA-Z0-9_.]+/)?(assets|data)(?:/[a-zA-Z0-9_.]+)+$"
)


class Extractor:
    """Base class for extractors."""

    directives: Dict[str, Directive]
    regex: Optional["re.Pattern[str]"]
    escaped_regex: Optional["re.Pattern[str]"]
    cache: Optional[Cache]

    def __init__(self, cache: Optional[Cache] = None):
        self.directives = {}
        self.regex = None
        self.escaped_regex = None
        self.cache = cache

    def generate_directives(self) -> str:
        names = list(self.directives)
        names.append("overlay")
        names.append("endoverlay")
        return "|".join(names)

    def generate_regex(self) -> str:
        """Return a regex that can match the current directives."""
        names = self.generate_directives()
        modifier = r"(?:\((?P<modifier>[^)]*)\)|\b)"
        arguments = r"(?P<arguments>.*)"
        return f"@(?P<name>{names}){modifier}{arguments}"

    def generate_escaped_regex(self) -> str:
        """Return a regex that can match escaped fragments."""
        names = self.generate_directives()
        return rf"(@@+(?:{names})\b.*)"

    def compile_regex(self, regex: str) -> "re.Pattern[str]":
        """Return the compiled pattern for the directive regex."""
        return re.compile(f"^{regex}$", flags=re.MULTILINE)

    def get_regex(self, directives: Mapping[str, Directive]) -> "re.Pattern[str]":
        """Create and return the regex for the specified directives."""
        directives = dict(directives)

        if self.regex is None or self.directives != directives:
            self.directives = directives
            self.regex = self.compile_regex(self.generate_regex())
            self.escaped_regex = self.compile_regex(self.generate_escaped_regex())

        return self.regex

    def split(
        self,
        source: str,
        directives: Mapping[str, Directive],
    ) -> Iterator[Tuple[str, Optional[Fragment]]]:
        """Split the given content into fragments."""
        lines = source.splitlines(keepends=True)
        previous_line = 0

        for fragment in self.parse_fragments(source, directives):
            if fragment.start_line > previous_line:
                yield "".join(lines[previous_line : fragment.start_line]), None
            yield "".join(lines[fragment.start_line : fragment.end_line]), fragment
            previous_line = fragment.end_line

        if previous_line < len(lines):
            yield "".join(lines[previous_line:]), None

    def extract(
        self,
        source: str,
        directives: Mapping[str, Directive],
        loaders: Iterable[FragmentLoader] = (),
    ) -> Tuple[ResourcePack, DataPack]:
        """Extract a resource pack and a data pack."""
        assets, data = ResourcePack(), DataPack()
        self.apply_directives(
            assets, data, directives, self.parse_fragments(source, directives), loaders
        )
        return assets, data

    def apply_directives(
        self,
        assets: ResourcePack,
        data: DataPack,
        directives: Mapping[str, Directive],
        fragments: Iterable[Fragment],
        loaders: Iterable[FragmentLoader] = (),
    ):
        """Apply directives into a blank data pack and a blank resource pack."""
        original_packs = assets, data
        added_overlays: Set[str] = set()

        for fragment in fragments:
            for loader in loaders:
                fragment = loader(fragment, directives)
                if not fragment:
                    break
            if fragment:
                if fragment.directive == "overlay":
                    directory = fragment.expect("directory")
                    formats: Any = (
                        json.loads(fragment.modifier) if fragment.modifier else {}
                    )
                    assets = assets.overlays.setdefault(directory, **formats)
                    data = data.overlays.setdefault(directory, **formats)
                    added_overlays.add(directory)
                elif fragment.directive == "endoverlay":
                    assets, data = original_packs
                else:
                    directives[fragment.directive](fragment, assets, data)

        for pack in original_packs:
            pack.overlays.cleanup(added_overlays)

    def parse_fragments(
        self,
        source: str,
        directives: Mapping[str, Directive],
    ) -> Iterator[Fragment]:
        """Parse and yield pack fragments."""
        return iter([])

    def create_fragment(
        self,
        start_line: int,
        end_line: int,
        match: "re.Match[str]",
    ):
        """Helper for creating a fragment from a matched pattern."""
        directive, modifier, arguments = match.groups()
        return Fragment(
            start_line=start_line,
            end_line=end_line,
            directive=directive,
            modifier=modifier,
            arguments=arguments.split(),
            cache=self.cache,
        )


class TextExtractor(Extractor):
    """Extractor for plain text files."""

    def parse_fragments(
        self,
        source: str,
        directives: Mapping[str, Directive],
    ) -> Iterator[Fragment]:
        tokens = self.get_regex(directives).split(source + "\n")
        assert self.escaped_regex

        it = iter(tokens)
        newlines = next(it).count("\n")

        while True:
            try:
                directive, modifier, arguments, content = islice(it, 4)
            except ValueError:
                return
            else:
                content = content.partition("\n")[-1]
                content = "".join(
                    s.replace("@@", "@", 1) if i % 2 else s
                    for i, s in enumerate(self.escaped_regex.split(content))
                )
                yield Fragment(
                    start_line=newlines,
                    end_line=(newlines := newlines + content.count("\n") + 1),
                    directive=directive,
                    modifier=modifier,
                    arguments=arguments.split(),
                    content=content[:-1] if content.endswith("\n") else content,
                    cache=self.cache,
                )


class EmbeddedExtractor(TextExtractor):
    """Extractor for directives embedded in markdown code blocks."""

    def generate_regex(self) -> str:
        return r"(?://|#)\s*" + super().generate_regex()

    def generate_escaped_regex(self) -> str:
        return r"(?://|#)\s*" + super().generate_escaped_regex()


class MarkdownParserWithUncheckedLinks(MarkdownIt):
    """MarkownIt subclass that removes link checks because lectern has its own logic."""

    def normalizeLink(self, url: str) -> str:
        return url

    def normalizeLinkText(self, link: str) -> str:
        return link

    def validateLink(self, url: str) -> bool:
        return True


class MarkdownExtractor(Extractor):
    """Extractor for markdown files."""

    text_extractor: TextExtractor
    embedded_extractor: TextExtractor
    comment_extractor: TextExtractor
    parser: MarkdownIt
    html_comment_regex: "re.Pattern[str]"
    embedded_sniffer: "re.Pattern[str]"

    def __init__(self, cache: Optional[Cache] = None):
        super().__init__(cache)
        self.text_extractor = TextExtractor(cache)
        self.embedded_extractor = EmbeddedExtractor(cache)
        self.comment_extractor = TextExtractor(cache)
        self.parser = MarkdownParserWithUncheckedLinks()
        self.html_comment_regex = re.compile(r"\s*<!--\s*(.+?)\s*-->\s*")
        self.embedded_sniffer = re.compile(
            r"^[ \t]*(?://|#)[ \t]*@[a-z0-9_]{3,}",
            re.MULTILINE,
        )

    def extract(
        self,
        source: str,
        directives: Mapping[str, Directive],
        loaders: Iterable[FragmentLoader] = (),
        external_files: Optional[FileSystemPath] = None,
    ) -> Tuple[ResourcePack, DataPack]:
        assets, data = ResourcePack(), DataPack()
        self.apply_directives(
            assets,
            data,
            directives,
            self.parse_fragments(source, directives, external_files),
            loaders,
        )
        return assets, data

    def parse_fragments(
        self,
        source: str,
        directives: Mapping[str, Directive],
        external_files: Optional[FileSystemPath] = None,
    ) -> Iterator[Fragment]:
        tokens = self.parser.parse(source)
        regex = self.get_regex(directives)

        skip_to = 0

        for i, token in enumerate(tokens):
            if not token.map:
                continue

            current_line = token.map[0]
            if skip_to > current_line:
                continue

            #
            # `@directive args...`
            #
            # ```
            # content
            # ```
            #
            if (
                skip_to := self.match_tokens(
                    tokens[i : i + 4],
                    "paragraph_open",
                    "inline",
                    "paragraph_close",
                    ["fence", "code_block"],
                )
            ) and (
                fragments := list(
                    self.parse_inline(tokens[i + 1], directives, external_files)
                )
            ):
                yield from fragments[:-1]
                yield replace(
                    fragments[-1].with_content(tokens[i + 3].content), end_line=skip_to
                )

            #
            # `@directive args...`
            #
            # ![](path/to/image)
            #
            elif (
                (
                    skip_to := self.match_tokens(
                        tokens[i : i + 6],
                        "paragraph_open",
                        "inline",
                        "paragraph_close",
                        "paragraph_open",
                        "inline",
                        "paragraph_close",
                    )
                )
                and (image := tokens[i + 4])
                and image.children
                and self.match_tokens(image.children, "image")
                and (link := image.children[0].attrGet("src"))
                and (
                    fragments := list(
                        self.parse_inline(tokens[i + 1], directives, external_files)
                    )
                )
            ):
                yield from fragments[:-1]
                yield replace(
                    fragments[-1].with_link(link, external_files), end_line=skip_to
                )

            #
            # `@directive args...`
            #
            # <details>
            #
            # ```
            # content
            # ```
            #
            # </details>
            #
            elif (
                (
                    skip_to := self.match_tokens(
                        tokens[i : i + 6],
                        "paragraph_open",
                        "inline",
                        "paragraph_close",
                        "html_block",
                        ["fence", "code_block"],
                        "html_block",
                    )
                )
                and tokens[i + 3].content == "<details>\n"
                and tokens[i + 5].content == "</details>\n"
                and (
                    fragments := list(
                        self.parse_inline(tokens[i + 1], directives, external_files)
                    )
                )
            ):
                yield from fragments[:-1]
                yield replace(
                    fragments[-1].with_content(tokens[i + 4].content), end_line=skip_to
                )

            #
            # `@directive args...`
            #
            # <details>
            #
            # ![](path/to/image)
            #
            # </details>
            #
            elif (
                (
                    skip_to := self.match_tokens(
                        tokens[i : i + 8],
                        "paragraph_open",
                        "inline",
                        "paragraph_close",
                        "html_block",
                        "paragraph_open",
                        "inline",
                        "paragraph_close",
                        "html_block",
                    )
                )
                and tokens[i + 3].content == "<details>\n"
                and tokens[i + 7].content == "</details>\n"
                and (image := tokens[i + 5])
                and image.children
                and self.match_tokens(image.children, "image")
                and (link := image.children[0].attrGet("src"))
                and (
                    fragments := list(
                        self.parse_inline(tokens[i + 1], directives, external_files)
                    )
                )
            ):
                yield from fragments[:-1]
                yield replace(
                    fragments[-1].with_link(link, external_files), end_line=skip_to
                )

            #
            # <!-- @directive args... -->
            #
            # ```
            # content
            # ```
            #
            elif (
                (
                    skip_to := self.match_tokens(
                        tokens[i : i + 2],
                        "html_block",
                        ["fence", "code_block"],
                    )
                )
                and (comment := self.html_comment_regex.match(token.content))
                and (match := regex.match(comment.group(1)))
            ):
                yield self.create_fragment(current_line, skip_to, match).with_content(
                    tokens[i + 1].content
                )

            #
            # <!-- @directive args... -->
            #
            # ![](path/to/image)
            #
            elif (
                (
                    skip_to := self.match_tokens(
                        tokens[i : i + 4],
                        "html_block",
                        "paragraph_open",
                        "inline",
                        "paragraph_close",
                    )
                )
                and (comment := self.html_comment_regex.match(token.content))
                and (image := tokens[i + 2])
                and image.children
                and self.match_tokens(image.children, "image")
                and (link := image.children[0].attrGet("src"))
                and (match := regex.match(comment.group(1)))
            ):
                yield self.create_fragment(current_line, skip_to, match).with_link(
                    link, external_files
                )

            #
            # [`@directive args...`](path/to/content)
            # `@directive args...`
            #
            elif (skip_to := self.match_tokens([token], "inline")) and (
                fragments := list(self.parse_inline(token, directives, external_files))
            ):
                yield from fragments

            #
            # <!-- @directive args... -->
            #
            elif (
                token.type == "html_block"
                and (comment := self.html_comment_regex.match(token.content))
                and (match := regex.match(comment.group(1)))
            ):
                yield self.create_fragment(current_line, skip_to, match)

            #
            # ```
            # @directive args...
            #
            # content
            # ```
            #
            elif token.type in ["fence", "code_block"]:
                offset = int(token.type == "fence")
                extractor = (
                    self.embedded_extractor
                    if self.embedded_sniffer.search(token.content)
                    else self.text_extractor
                )
                for fragment in extractor.parse_fragments(token.content, directives):
                    yield replace(
                        fragment,
                        start_line=fragment.start_line + current_line + offset,
                        end_line=(skip_to := fragment.end_line + current_line),
                    )

            #
            # <!--
            # @directive args...
            #
            # content
            # -->
            #
            elif (
                token.type == "html_block"
                and (html := token.content.rstrip())
                and html.startswith("<!--")
                and html.endswith("-->")
            ):
                for fragment in self.comment_extractor.parse_fragments(
                    html[4:-3],
                    directives,
                ):
                    yield replace(
                        fragment,
                        start_line=fragment.start_line + current_line,
                        end_line=(skip_to := fragment.end_line + current_line),
                    )

    def parse_inline(
        self,
        token: Token,
        directives: Mapping[str, Directive],
        external_files: Optional[FileSystemPath] = None,
    ) -> Iterator[Fragment]:
        if token.type == "inline" and token.children and token.map:
            regex = self.get_regex(directives)

            it = iter(enumerate(token.children))

            for i, child in it:
                if (
                    self.match_newline(token.children, i - 1)
                    and self.match_newline(token.children, i + 1)
                    and child.type == "code_inline"
                ):
                    code = child
                elif (
                    self.match_newline(token.children, i - 1)
                    and self.match_newline(token.children, i + 3)
                    and self.match_tokens(
                        token.children[i : i + 3],
                        "link_open",
                        "code_inline",
                        "link_close",
                    )
                ):
                    code = token.children[i + 1]
                    next(it)
                    next(it)
                else:
                    continue

                if match := (
                    regex.match(code.content)
                    or (
                        (directory := RELATIVE_PATH_REGEX.match(code.content))
                        and (
                            directive := (
                                "@resource_pack"
                                if directory[0] == "assets"
                                else "@data_pack"
                            )
                        )
                        and regex.match(f"{directive} {code.content}")
                    )
                ):
                    start_line, end_line = token.map
                    fragment = self.create_fragment(start_line, end_line, match)
                    if child.type == "link_open" and (link := child.attrGet("href")):
                        fragment = fragment.with_link(link, external_files)
                    yield fragment

    def match_tokens(
        self,
        tokens: Optional[List[Token]],
        *token_types: Union[List[str], str],
    ) -> int:
        """Return whether the list of tokens matches the provided token types."""
        return (
            tokens is not None
            and len(tokens) == len(token_types)
            and all(
                (
                    token_type == token.type
                    if isinstance(token_type, str)
                    else token.type in token_type
                )
                for token, token_type in zip(tokens, token_types)
            )
            and next((token.map[-1] for token in reversed(tokens) if token.map), 1)
        )

    def match_newline(self, tokens: Optional[List[Token]], index: int) -> bool:
        return (
            not tokens
            or index < 0
            or index >= len(tokens)
            or tokens[index].type in ("softbreak", "hardbreak")
        )



"""Plugin that adds a directive for rendering a fragment and storing the output."""

__all__ = [
    "DefineDirective",
]


from dataclasses import dataclass

from beet import Context, DataPack, ResourcePack, TextFile

from lectern import Document, Fragment


def beet_default(ctx: Context):
    document = ctx.inject(Document)
    document.directives["define"] = DefineDirective(ctx)


@dataclass
class DefineDirective:
    """Directive that renders the fragment and stores the resulting string as a template global."""

    ctx: Context

    def __call__(self, fragment: Fragment, assets: ResourcePack, data: DataPack):
        variable = fragment.expect("variable")
        file_instance = fragment.as_file(TextFile)

        result = self.ctx.template.render_file(file_instance)
        self.ctx.template.globals[variable] = result.text



__all__ = [
    "TextSerializer",
    "MarkdownSerializer",
    "ExternalFilesManager",
    "SerializedFile",
    "snapshot_serialization_filter",
    "EXTENSION_HIGHLIGHTING",
]

import os
import re
from base64 import b64encode
from contextlib import contextmanager
from dataclasses import dataclass, field
from itertools import chain
from mimetypes import guess_type
from pathlib import Path
from textwrap import indent
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Type,
    Union,
)
from urllib.parse import urlparse

from beet import (
    DataPack,
    File,
    NamespaceFile,
    ResourcePack,
    TextFile,
    TextFileBase,
    get_output_scope,
)
from beet.core.utils import normalize_string

EXTENSION_HIGHLIGHTING = {
    ".mcfunction": "mcfunction",
    ".json": "json",
    ".mcmeta": "json",
    ".fsh": "glsl",
    ".vsh": "glsl",
    ".glsl": "glsl",
}


def snapshot_serialization_filter(pack: Union[DataPack, ResourcePack]) -> bool:
    """Filter that selects non-empty packs and empty packs with modified pack.mcmeta."""
    return bool(pack or pack.mcmeta != type(pack)().mcmeta)


@dataclass
class ExternalFilesManager:
    """Class responsible for saving external files."""

    directory: Path
    target: Path

    external_files: Dict[str, File[Any, Any]] = field(default_factory=dict)
    external_prefix: str = ""

    def __post_init__(self):
        if not self.external_prefix:
            prefix = os.path.relpath(self.directory, self.target.parent)
            prefix = prefix.replace(os.path.sep, "/")
            self.external_prefix = "" if prefix == "." else prefix.rstrip("/") + "/"

    def flush(self):
        """Save all the external files."""
        if self.external_files:
            self.directory.mkdir(parents=True, exist_ok=True)
            for filename, file_instance in self.external_files.items():
                file_instance.dump(self.directory, self.target.parent / filename)

    def __enter__(self) -> "ExternalFilesManager":
        return self

    def __exit__(self, *_):
        self.flush()


@dataclass
class SerializedFile:
    """Class representing a file emitted during the serialization process."""

    hint: str
    file_instance: File[Any, Any]

    filename: str = field(init=False)
    extension: str = field(init=False)
    content_type: str = field(init=False)

    def __post_init__(self):
        path = Path(urlparse(self.hint).path)
        self.extension = "".join(path.suffixes)
        self.filename = normalize_string(
            path.name[: -len(self.extension)] if self.extension else path.name
        )

        if not self.extension:
            self.extension = getattr(self.file_instance, "extension", self.extension)

        self.content_type = (
            guess_type(f"{self.filename}{self.extension}")[0]
            or "application/octet-stream"
        )

    def get_local_path(
        self,
        external_files: Dict[str, File[Any, Any]],
        external_prefix: str = "",
    ) -> str:
        """Rename the file to prevent conflicts within the given environment."""
        while (
            path := f"{external_prefix}{self.filename}{self.extension}"
        ) in external_files:
            stem, _, number = self.filename.rpartition("_")
            if not number.isdigit():
                stem, number = self.filename, "0"
            self.filename = f"{stem}_{int(number) + 1}"
        return path

    def get_data_url(self):
        """Convert the file to a data url."""
        content = self.file_instance.ensure_serialized()
        return f"data:{self.content_type};base64,{b64encode(content).decode()}"

    def generate_link_target(
        self,
        external_files: Optional[Dict[str, File[Any, Any]]] = None,
        external_prefix: str = "",
    ) -> str:
        """Generate the link for including the file in a serialized document."""
        if external_files is None:
            return self.get_data_url()

        local_path = self.get_local_path(external_files, external_prefix)
        external_files[local_path] = self.file_instance
        return local_path


class TextSerializer:
    """Document serializer that outputs plain text."""

    sort: bool
    escaped_regex: Optional["re.Pattern[str]"]
    pack_filter: Callable[[Union[ResourcePack, DataPack]], bool]

    def __init__(self, *, sort: bool = False):
        self.sort = sort
        self.escaped_regex = None
        self.pack_filter = bool

    def sort_items[K, V](self, obj: Mapping[K, V]) -> Iterable[tuple[K, V]]:
        return (
            sorted(obj.items(), key=lambda p: str(p[0])) if self.sort else obj.items()
        )

    @contextmanager
    def use_sorted_items(self, sort: bool = True):
        """Context manager for sorting items."""
        previous = self.sort
        self.sort = sort

        try:
            yield
        finally:
            self.sort = previous

    def get_escaped_regex(
        self, mapping: Mapping[Type[NamespaceFile], str]
    ) -> "re.Pattern[str]":
        """Create and return the escaped regex for the specified serialization mapping."""
        names = "|".join([*mapping.values(), "resource_pack", "data_pack"])
        pattern = rf"^(@+(?:{names})\b.*)$"

        if self.escaped_regex is None or self.escaped_regex.pattern != pattern:
            self.escaped_regex = re.compile(pattern, flags=re.MULTILINE)

        return self.escaped_regex

    def serialize(
        self,
        assets: ResourcePack,
        data: DataPack,
        mapping: Mapping[Type[NamespaceFile], str],
    ) -> str:
        """Return the serialized representation."""
        result: List[str] = []

        for pack, extra_directive in [
            (assets, "resource_pack"),
            (data, "data_pack"),
        ]:
            if not self.pack_filter(pack):
                continue

            result.append(self.serialize_pack(pack, extra_directive, mapping))

            if pack.overlay_parent is None:
                should_end = False
                for overlay in pack.overlays.values():
                    if overlay:
                        should_end = True
                        result.append(
                            self.serialize_pack(overlay, extra_directive, mapping)
                        )
                if should_end:
                    result.append("@endoverlay\n")

        return "\n".join(result)

    def serialize_pack(
        self,
        pack: Union[DataPack, ResourcePack],
        extra_directive: str,
        mapping: Mapping[Type[NamespaceFile], str],
    ) -> str:
        escaped_regex = self.get_escaped_regex(mapping)

        return "\n".join(
            (
                f"@{directive_name} {argument}\n{self.escape_fragment(content, escaped_regex)}"
                if isinstance(content := file_instance.ensure_serialized(), str)
                else f"@{directive_name}(base64) {argument}\n"
                + b64encode(content).decode()
                + "\n"
            )
            for directive_name, argument, file_instance in chain(
                (
                    [("overlay", pack.overlay_name, TextFile())]
                    if pack.overlay_name is not None
                    else []
                ),
                (
                    (extra_directive, path, file_instance)
                    for path, file_instance in self.sort_items(pack.extra)
                    if not (
                        pack.overlay_parent is not None
                        and path in ("pack.mcmeta", "pack.png")
                    )
                ),
                (
                    (
                        extra_directive,
                        f"{namespace.directory}/{name}/{path}",
                        file_instance,
                    )
                    for name, namespace in self.sort_items(pack)
                    for path, file_instance in self.sort_items(namespace.extra)
                ),
                (
                    (
                        (
                            mapping[file_type],
                            f"{name}:{path}",
                            file_instance,
                        )
                        if file_type in mapping
                        else (
                            extra_directive,
                            f"{namespace.directory}/{name}/{'/'.join(get_output_scope(file_type.scope, pack.pack_format))}/{path}{file_type.extension}",
                            file_instance,
                        )
                    )
                    for name, namespace in self.sort_items(pack)
                    for file_type, container in self.sort_items(namespace)
                    for path, file_instance in self.sort_items(container)
                ),
            )
        )

    def escape_fragment(self, content: str, escaped_regex: "re.Pattern[str]") -> str:
        """Escape embedded directives."""
        return "".join(
            s.replace("@", "@@", 1) if i % 2 else s
            for i, s in enumerate(escaped_regex.split(content))
        )


class MarkdownSerializer:
    """Document serializer that outputs markdown and emits associated files."""

    sort: bool
    flat: bool
    pack_filter: Callable[[Union[ResourcePack, DataPack]], bool]

    def __init__(self, *, sort: bool = False, flat: bool = False):
        self.sort = sort
        self.flat = flat
        self.pack_filter = bool

    def sort_items[K, V](self, obj: Mapping[K, V]) -> Iterable[tuple[K, V]]:
        return (
            sorted(obj.items(), key=lambda p: str(p[0])) if self.sort else obj.items()
        )

    @contextmanager
    def use_sorted_items(self, sort: bool = True):
        """Context manager for sorting items."""
        previous = self.sort
        self.sort = sort

        try:
            yield
        finally:
            self.sort = previous

    @contextmanager
    def use_flat_format(self, flat: bool = True):
        """Context manager for using the flat markdown format."""
        previous = self.flat
        self.flat = flat

        try:
            yield
        finally:
            self.flat = previous

    def serialize(
        self,
        assets: ResourcePack,
        data: DataPack,
        mapping: Mapping[Type[NamespaceFile], str],
        external_files: Optional[Dict[str, File[Any, Any]]] = None,
        external_prefix: str = "",
    ) -> str:
        """Return the serialized representation and the files emitted in the process."""
        return "# Lectern snapshot\n\n" + (
            "\n".join(
                text
                for title, directive, pack in [
                    ("Data pack", "data_pack", data),
                    ("Resource pack", "resource_pack", assets),
                ]
                if self.pack_filter(pack)
                for text in self.serialize_pack(
                    title,
                    directive,
                    pack,
                    mapping,
                    external_files,
                    external_prefix,
                )
            )
            or "The data pack and resource pack are empty.\n"
        )

    def serialize_pack(
        self,
        title: str,
        pack_directive: str,
        pack: Union[DataPack, ResourcePack],
        mapping: Mapping[Type[NamespaceFile], str],
        external_files: Optional[Dict[str, File[Any, Any]]] = None,
        external_prefix: str = "",
    ) -> Iterator[str]:
        """Yield markdown chunks for the given pack."""
        yield f"## {title}"

        if pack.overlay_name is not None:
            yield f"\n`@overlay {pack.overlay_name}`"

        for path, file_instance in self.sort_items(pack.extra):
            if pack.overlay_parent is not None and path in ("pack.mcmeta", "pack.png"):
                continue
            yield from self.format_serialized_file(
                self.serialize_file_instance(
                    pack_directive,
                    path,
                    file_instance,
                    external_files,
                    external_prefix,
                )
            )

        for name, namespace in self.sort_items(pack):
            if not namespace:
                continue

            yield f"\n### {name}"

            for path, file_instance in self.sort_items(namespace.extra):
                yield from self.format_serialized_file(
                    self.serialize_file_instance(
                        pack_directive,
                        f"{namespace.directory}/{name}/{path}",
                        file_instance,
                        external_files,
                        external_prefix,
                    )
                )

            for file_type, container in self.sort_items(namespace):
                for path, file_instance in self.sort_items(container):
                    if file_type in mapping:
                        yield from self.format_serialized_file(
                            self.serialize_file_instance(
                                mapping[file_type],
                                f"{name}:{path}",
                                file_instance,  # pyright: ignore[reportArgumentType]
                                external_files,
                                external_prefix,
                            )
                        )
                    else:
                        yield from self.format_serialized_file(
                            self.serialize_file_instance(
                                pack_directive,
                                f"{namespace.directory}/{name}/{'/'.join(get_output_scope(file_type.scope, pack.pack_format))}/{path}{file_type.extension}",
                                file_instance,  # pyright: ignore[reportArgumentType]
                                external_files,
                                external_prefix,
                            )
                        )
        yield ""

        if pack.overlay_parent is None:
            should_end = False
            for directory, overlay in self.sort_items(pack.overlays):
                if overlay:
                    should_end = True
                    yield from self.serialize_pack(
                        f"Overlay `{directory}`",
                        pack_directive,
                        overlay,
                        mapping,
                        external_files,
                        external_prefix,
                    )
            if should_end:
                yield "`@endoverlay`\n"

    def format_serialized_file(self, chunks: Iterable[str]) -> Iterator[str]:
        """Format the markdown chunks for serializing file instances."""
        if self.flat:
            yield ""
        else:
            chunks = [indent(chunk, "  ") for chunk in chunks]
            chunks[0] = "\n-" + chunks[0][1:]
        yield from chunks

    def serialize_file_instance(
        self,
        directive: str,
        argument: str,
        file_instance: File[Any, Any],
        external_files: Optional[Dict[str, File[Any, Any]]] = None,
        external_prefix: str = "",
    ) -> Iterator[str]:
        """Yield markdown chunks for including the given file instance."""
        serialized_file = SerializedFile(argument, file_instance)

        if isinstance(file_instance, TextFileBase):
            content = file_instance.text

            if not content.endswith("\n"):
                directive += "(strip_final_newline)"
                content += "\n"

            yield f"`@{directive} {argument}`"
            if not self.flat:
                yield "\n<details>"
            yield "\n```" + EXTENSION_HIGHLIGHTING.get(serialized_file.extension, "")
            yield content + "```"
            if not self.flat:
                yield "\n</details>"

            return

        target = serialized_file.generate_link_target(external_files, external_prefix)

        if serialized_file.content_type.startswith("image/"):
            yield f"`@{directive} {argument}`"
            if not self.flat:
                yield "\n<details>"
            yield f"\n![{directive}{serialized_file.extension}]({target})"
            if not self.flat:
                yield "\n</details>"
        else:
            yield f"[`@{directive} {argument}`]({target})"



__all__ = [
    "beet_default",
    "lectern",
]


import subprocess
from glob import glob
from typing import List, Literal, Optional

from beet import Context, ListOption, PackageablePath, configurable
from pydantic import BaseModel

from .document import Document
from .loaders import LinkFragmentLoader


class LecternOptions(BaseModel):
    load: ListOption[PackageablePath] = ListOption()
    links: Literal["enable", "ignore", "disable"] = "enable"
    snapshot: Optional[str] = None
    snapshot_sort: bool = False
    snapshot_flat: bool = False
    external_files: Optional[PackageablePath] = None
    scripts: List[List[str]] = []


def beet_default(ctx: Context):
    ctx.require(lectern)


@configurable(validator=LecternOptions)
def lectern(ctx: Context, opts: LecternOptions):
    """Plugin that handles markdown files with lectern."""
    document = ctx.inject(Document)
    document.loaders.append(LinkFragmentLoader(status=opts.links))

    for pattern in opts.load.entries():
        for path in sorted(glob(str(ctx.directory / pattern))):
            document.load(path)

    for arguments in opts.scripts:
        result = subprocess.run(
            arguments,
            cwd=ctx.directory,
            check=True,
            stdout=subprocess.PIPE,
        )
        document.add_text(result.stdout.decode())

    yield

    if opts.snapshot:
        with (
            document.text_serializer.use_sorted_items(opts.snapshot_sort),
            document.markdown_serializer.use_sorted_items(opts.snapshot_sort),
            document.markdown_serializer.use_flat_format(opts.snapshot_flat),
        ):
            document.save(
                ctx.directory / opts.snapshot,
                ctx.directory / opts.external_files if opts.external_files else None,
            )



"""Plugin that adds a directive for interpreting lectern text templates."""

__all__ = [
    "ScriptDirective",
]


from dataclasses import dataclass

from beet import Context, DataPack, ResourcePack, TextFile

from lectern import Document, Fragment


def beet_default(ctx: Context):
    document = ctx.inject(Document)
    document.directives["script"] = ScriptDirective(ctx)


@dataclass
class ScriptDirective:
    """Directive that renders the fragment and interprets the result as lectern text."""

    ctx: Context

    def __call__(self, fragment: Fragment, assets: ResourcePack, data: DataPack):
        fragment.expect()
        file_instance = fragment.as_file(TextFile)

        source = self.ctx.template.render_file(file_instance)

        document = self.ctx.inject(Document)
        document.add_text(source.text)



"""Plugin for handling markdown coming from messaging apps."""

__all__ = [
    "MessagingOptions",
    "messaging",
]


from typing import Literal, Optional

from beet import Context, ErrorMessage, Function, ListOption, configurable
from pydantic import BaseModel

from lectern import Document, LinkFragmentLoader


class MessagingOptions(BaseModel):
    links: Literal["enable", "ignore", "disable"] = "enable"
    input: ListOption[str] = ListOption()
    nothing_error: Optional[str] = None


def beet_default(ctx: Context):
    ctx.require(messaging)


@configurable(validator=MessagingOptions)
def messaging(ctx: Context, opts: MessagingOptions):
    """Plugin for handling markdown coming from messaging apps.

    The extractor cache is disabled to prevent remote files from being
    written to disk. If the message doesn't contain any fragment we try to
    form a default function by concatenating all code blocks.
    """
    document = ctx.inject(Document)
    document.loaders.append(LinkFragmentLoader(status=opts.links))
    document.markdown_extractor.cache = None

    directives = document.directives.resolve()

    for message in opts.input.entries():
        if fragments := list(
            document.markdown_extractor.parse_fragments(message, directives)
        ):
            document.markdown_extractor.apply_directives(
                ctx.assets, ctx.data, directives, fragments, document.loaders
            )
        elif code_blocks := [
            token.content
            for token in document.markdown_extractor.parser.parse(message)
            if token.type in ["fence", "code_block"]
        ]:
            ctx.generate("default", Function("\n".join(code_blocks)))
        elif "```" in message:
            raise ErrorMessage(
                "Couldn't parse code block. Make sure the triple backticks are on their own line."
            )
        elif opts.nothing_error:
            raise ErrorMessage(opts.nothing_error)



"""Plugin that adds a directive for running plugin code in a document."""

__all__ = [
    "PluginDirective",
]


from dataclasses import dataclass
from textwrap import indent

from beet import Context, DataPack, ResourcePack, TextFile
from beet.core.utils import JsonDict

from lectern import Document, Fragment


def beet_default(ctx: Context):
    document = ctx.inject(Document)
    document.directives["plugin"] = PluginDirective(ctx)


@dataclass
class PluginDirective:
    """Directive that evaluates the fragment as a beet plugin."""

    ctx: Context

    def __call__(self, fragment: Fragment, assets: ResourcePack, data: DataPack):
        file_instance = fragment.as_file(TextFile)
        body = indent(file_instance.text.strip() or "pass", "    ")
        globals_dict: JsonDict = {}
        exec("def beet_default(ctx):\n" + body, globals_dict)
        self.ctx.require(globals_dict["beet_default"])



__all__ = [
    "Directive",
    "DirectiveRegistry",
    "NamespacedResourceDirective",
    "DataPackDirective",
    "ResourcePackDirective",
    "BundleFragmentMixin",
    "MergeZipDirective",
    "SkipDirective",
]


import io
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Protocol, Type
from zipfile import ZipFile

from beet import Container, DataPack, NamespaceFile, ResourcePack
from beet.core.utils import snake_case

from .fragment import Fragment


class Directive(Protocol):
    """Protocol for detecting directives."""

    def __call__(self, fragment: Fragment, assets: ResourcePack, data: DataPack, /): ...


class DirectiveRegistry(Container[str, Directive]):
    """Registry for directives."""

    resolvers: List[Callable[["DirectiveRegistry"], Any]]
    assets: ResourcePack
    data: DataPack

    def __init__(
        self,
        assets: Optional[ResourcePack] = None,
        data: Optional[DataPack] = None,
    ):
        super().__init__()
        self.resolvers = []
        self.assets = assets or ResourcePack()
        self.data = data or DataPack()

        self["data_pack"] = DataPackDirective()
        self["resource_pack"] = ResourcePackDirective()
        self["merge_zip"] = MergeZipDirective()
        self["skip"] = SkipDirective()

        @self.add_resolver
        def _(self: DirectiveRegistry):
            for pack in [self.assets, self.data]:
                for file_type in pack.resolve_scope_map().values():
                    name = snake_case(file_type.__name__)
                    self[name] = NamespacedResourceDirective(file_type)

    def resolve(self) -> "DirectiveRegistry":
        """Resolve all directives in the registry."""
        for callback in self.resolvers:
            callback(self)
        return self

    def add_resolver(self, resolver: Callable[["DirectiveRegistry"], Any]):
        """Add directive resolver."""
        self.resolvers.append(resolver)

    def get_serialization_mapping(self) -> Dict[Type[NamespaceFile], str]:
        """Return the serialization mapping."""
        return {
            directive.file_type: directive_name
            for directive_name, directive in self.items()
            if isinstance(directive, NamespacedResourceDirective)
        }


@dataclass
class NamespacedResourceDirective:
    """Directive for including namespaced resources."""

    file_type: Type[NamespaceFile]

    def __call__(self, fragment: Fragment, assets: ResourcePack, data: DataPack):
        full_name = fragment.expect("full_name")
        file_instance: Any = fragment.as_file(self.file_type)  # pyright: ignore[reportArgumentType]

        pack = assets if self.file_type in assets.namespace_type.field_map else data
        proxy: Any = pack[self.file_type]

        if fragment.modifier == "append":
            current_file = proxy.setdefault(full_name, self.file_type(""))
            current_file.text += file_instance.text
        elif fragment.modifier == "prepend":
            current_file = proxy.setdefault(full_name, self.file_type(""))
            current_file.text = file_instance.text + current_file.text
        elif fragment.modifier == "merge":
            proxy.merge({full_name: file_instance})
        else:
            proxy[full_name] = file_instance


class BundleFragmentMixin:
    """Directive mixin that can bundle a fragment into a zipfile."""

    def bundle_pack_fragment(self, fragment: Fragment) -> ZipFile:
        """Return a zipfile containing the pack fragment."""
        relative_path = fragment.expect("relative_path")
        file_instance = fragment.as_file()

        data = io.BytesIO()
        with ZipFile(data, mode="w") as zip_file:
            file_instance.dump(zip_file, relative_path)

        data.seek(0)
        return ZipFile(data)


@dataclass
class DataPackDirective(BundleFragmentMixin):
    """Directive that loads the fragment into the data pack."""

    def __call__(self, fragment: Fragment, assets: ResourcePack, data: DataPack):
        data.load(self.bundle_pack_fragment(fragment))


@dataclass
class ResourcePackDirective(BundleFragmentMixin):
    """Directive that loads the fragment into the resource pack."""

    def __call__(self, fragment: Fragment, assets: ResourcePack, data: DataPack):
        assets.load(self.bundle_pack_fragment(fragment))


@dataclass
class MergeZipDirective:
    """Directive that merges a zipped resource pack or data pack."""

    def __call__(self, fragment: Fragment, assets: ResourcePack, data: DataPack):
        fragment.expect()
        file_instance = fragment.as_file()
        with ZipFile(io.BytesIO(file_instance.blob)) as origin:
            for path in origin.namelist():
                if path.startswith("assets/"):
                    assets.load(origin)
                    break
                elif path.startswith("data/"):
                    data.load(origin)
                    break


@dataclass
class SkipDirective:
    """Directive that ignores the fragment and skips to the next one."""

    def __call__(self, fragment: Fragment, assets: ResourcePack, data: DataPack):
        fragment.expect()



__all__ = [
    "LinkFragmentLoader",
]


from dataclasses import dataclass
from typing import Literal, Mapping, Optional

from .directive import Directive
from .fragment import Fragment, InvalidFragment


@dataclass
class LinkFragmentLoader:
    """Loader for validating link fragments."""

    status: Literal["enable", "ignore", "disable"] = "enable"

    def __call__(
        self,
        fragment: Fragment,
        directives: Mapping[str, Directive],
    ) -> Optional[Fragment]:
        if fragment.url or fragment.modifier == "download":
            if self.status == "ignore":
                return None
            if self.status == "disable":
                msg = "Link fragments are disabled."
                raise InvalidFragment(msg, fragment.start_line)
        return fragment



"""Plugin that adds a directive for requiring plugins dynamically."""

__all__ = [
    "RequireDirective",
]


from dataclasses import dataclass

from beet import Context, DataPack, ResourcePack

from lectern import Document, Fragment


def beet_default(ctx: Context):
    document = ctx.inject(Document)
    document.directives["require"] = RequireDirective(ctx)


@dataclass
class RequireDirective:
    """Directive that requires a given plugin."""

    ctx: Context

    def __call__(self, fragment: Fragment, assets: ResourcePack, data: DataPack):
        plugin = fragment.expect("plugin")

        self.ctx.require(plugin)



__version__ = "0.35.0-beta.6"

from .directive import *  # noqa: F403
from .document import *  # noqa: F403
from .extract import *  # noqa: F403
from .fragment import *  # noqa: F403
from .loaders import *  # noqa: F403
from .plugin import *  # noqa: F403
from .prefetch import *  # noqa: F403
from .serialize import *  # noqa: F403



__all__ = [
    "tagged_function_shorthand",
    "TaggedFunctionDirective",
]


from dataclasses import dataclass, replace
from typing import Dict, Sequence, Type

from beet import (
    Context,
    DataPack,
    Function,
    Generator,
    ListOption,
    NamespaceFile,
    PluginOptions,
    ResourcePack,
    configurable,
)
from beet.core.utils import required_field

from lectern import Document, Fragment, NamespacedResourceDirective


class TaggedFunctionOptions(PluginOptions):
    tags: Dict[str, ListOption[str]] = {}


def beet_default(ctx: Context):
    ctx.require(tagged_function_shorthand)


@configurable(validator=TaggedFunctionOptions)
def tagged_function_shorthand(ctx: Context, opts: TaggedFunctionOptions):
    document = ctx.inject(Document)
    for shorthand, tags in opts.tags.items():
        document.directives[f"{shorthand}_function"] = TaggedFunctionDirective(
            shorthand=shorthand,
            tags=tags.entries(),
            generate=ctx.generate,
        )


@dataclass
class TaggedFunctionDirective(NamespacedResourceDirective):
    shorthand: str = required_field()
    tags: Sequence[str] = required_field()
    generate: Generator = required_field()

    file_type: Type[NamespaceFile] = Function

    def __call__(self, fragment: Fragment, assets: ResourcePack, data: DataPack):
        if not fragment.arguments:
            fragment = replace(
                fragment, arguments=(self.generate.path(self.shorthand),)
            )

        super().__call__(fragment, assets, data)

        full_name = fragment.expect("full_name")

        for tag in self.tags:
            data.function_tags.setdefault(tag).add(full_name)



__all__ = [
    "VanillaFilesDirective",
    "VanillaMatchDirective",
]


from dataclasses import dataclass
from typing import Union

from beet import Context, DataPack, PackQuery, ResourcePack
from beet.contrib.vanilla import Vanilla

from lectern import Document, Fragment, InvalidFragment


def beet_default(ctx: Context):
    vanilla = ctx.inject(Vanilla)
    document = ctx.inject(Document)
    document.directives["vanilla"] = VanillaFilesDirective(vanilla, ctx.query)
    for file_type in ctx.get_file_types():
        document.directives[f"vanilla_{file_type.snake_name}"] = VanillaMatchDirective(
            file_type.snake_name, vanilla, ctx.query
        )


@dataclass
class VanillaFilesDirective:
    vanilla: Vanilla
    query: PackQuery[Union[ResourcePack, DataPack]]

    def __call__(self, fragment: Fragment, assets: ResourcePack, data: DataPack):
        release = self.vanilla.releases[
            fragment.modifier or self.vanilla.minecraft_version
        ]

        if not fragment.arguments:
            raise InvalidFragment(
                "Missing arguments for directive @vanilla.", fragment.start_line
            )

        head, *tail = fragment.arguments

        client_jar = release.client_jar
        query = self.query.from_pack(client_jar.assets, client_jar.data).prepare(
            files={head[:-1]: tail} if head.endswith(":") else fragment.arguments
        )

        for base_path in query.analyze_base_paths():
            release.mount(base_path, fetch_objects=True)

        query.copy_to(assets, data)


@dataclass
class VanillaMatchDirective:
    group: str
    vanilla: Vanilla
    query: PackQuery[Union[ResourcePack, DataPack]]

    def __call__(self, fragment: Fragment, assets: ResourcePack, data: DataPack):
        release = self.vanilla.releases[
            fragment.modifier or self.vanilla.minecraft_version
        ]

        args = fragment.arguments or ["*"]
        head, *tail = args

        client_jar = release.client_jar
        query = self.query.from_pack(client_jar.assets, client_jar.data).prepare(
            match={self.group: {head[:-1]: tail} if head.endswith(":") else args}
        )

        for base_path in query.analyze_base_paths():
            release.mount(base_path, fetch_objects=True)

        query.copy_to(assets, data)



"""Plugin that handles relative resource locations."""

__all__ = [
    "RelativeNamespacedResourceLoader",
]


from dataclasses import dataclass, replace
from typing import Mapping

from beet import Context

from lectern import Directive, Document, Fragment, NamespacedResourceDirective


def beet_default(ctx: Context):
    document = ctx.inject(Document)
    document.loaders.append(RelativeNamespacedResourceLoader(ctx))


@dataclass
class RelativeNamespacedResourceLoader:
    """Loader that resolves relative resource locations."""

    ctx: Context

    def __call__(
        self,
        fragment: Fragment,
        directives: Mapping[str, Directive],
    ) -> Fragment:
        if (
            isinstance(directives.get(fragment.directive), NamespacedResourceDirective)
            and fragment.arguments
        ):
            name = fragment.arguments[0]
            if ":" not in name:
                fragment = replace(fragment, arguments=[self.ctx.generate.path(name)])
        return fragment
